import numpy as np
import numpy.random as rnd

from .Result import Result
from .State import State  # pylint: disable=unused-import
from .Statistics import Statistics
from .WeigthIndex import WeightIndex
from .criteria import AcceptanceCriterion  # pylint: disable=unused-import


class ALNS:

    def __init__(self, rnd_state=rnd.RandomState()):
        """
        Implements the adaptive large neighbourhood search (ALNS) algorithm.
        The implementation optimises for a minimisation problem, as explained
        in the text by Pisinger and Røpke (2010).

        Parameters
        ----------
        rnd_state : rnd.RandomState
            Optional random state to use for random number generation. When
            passed, this state is used for operator selection and general
            computations requiring random numbers. It is also passed to the
            destroy and repair operators, as a second argument.

        References
        ----------
        - Pisinger, D., and Røpke, S. (2010). Large Neighborhood Search. In M.
          Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399-420).
          Springer.
        """
        self._destroy_operators = []
        self._repair_operators = []

        self._rnd_state = rnd_state

    @property
    def destroy_operators(self):
        """
        Returns the destroy operators set for the ALNS algorithm.
        """
        return self._destroy_operators

    @property
    def repair_operators(self):
        """
        Returns the repair operators set for the ALNS algorithm.
        """
        return self._repair_operators

    def add_destroy_operator(self, operator):
        """
        Adds a destroy operator to the heuristic instance.

        Parameters
        ----------
        operator : Callable[[State, RandomState], State]
            An operator that, when applied to the current state, returns a new
            state reflecting its implemented destroy action. The second argument
            is the random state constructed from the passed-in seed.
        """
        self._destroy_operators.append(operator)

    def add_repair_operator(self, operator):
        """
        Adds a repair operator to the heuristic instance.

        Parameters
        ----------
        operator : Callable[[State, RandomState], State]
            An operator that, when applied to the destroyed state, returns a
            new state reflecting its implemented repair action. The second
            argument is the random state constructed from the passed-in seed.
        """
        self._repair_operators.append(operator)

    def iterate(self, initial_solution, weights, operator_decay, criterion,
                iterations=10000, collect_stats=True):
        """
        Runs the adaptive large neighbourhood search heuristic [1], using the
        previously set destroy and repair operators. The first solution is set
        to the passed-in initial solution, and then subsequent solutions are
        computed by iteratively applying the operators.

        Parameters
        ----------
        initial_solution : State
            The initial solution, as a State object.
        weights: array_like
            A list of four positive elements, representing the weight updates
            when the candidate solution results in a new global best (idx 0),
            is better than the current solution (idx 1), the solution is
            accepted (idx 2), or rejected (idx 3).
        operator_decay : float
            The operator decay parameter, as a float in the unit interval.
        criterion : AcceptanceCriterion
            The acceptance criterion to use for candidate states. See also
            the `alns.criteria` module for an overview.
        iterations : int
            The number of iterations. Default 10000.
        collect_stats : bool
            Should statistics be collected during iteration? Default True, but
            may be turned off for long runs to reduce memory consumption.

        Raises
        ------
        ValueError
            When the parameters do not meet requirements.

        Returns
        -------
        Result
            A result object, containing the best and last solutions, and some
            additional results.

        References
        ----------
        [1]: Pisinger, D., & Røpke, S. (2010). Large Neighborhood Search. In M.
        Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399-420).
        Springer.

        [2]: S. Røpke and D. Pisinger (2006). A unified heuristic for a large
        class of vehicle routing problems with backhauls. *European Journal of
        Operational Research*, 171: 750–775, 2006.
        """
        weights = np.asarray(weights, dtype=np.float16)

        self._validate_parameters(weights, operator_decay, iterations)

        current = best = initial_solution

        d_weights = np.ones_like(self.destroy_operators, dtype=np.float16)
        r_weights = np.ones_like(self.repair_operators, dtype=np.float16)

        statistics = Statistics()

        if collect_stats:
            statistics.collect_objective(initial_solution.objective())

        for iteration in range(iterations):
            d_idx = select_operator(self.destroy_operators, d_weights,
                                    self._rnd_state)

            r_idx = select_operator(self.repair_operators, r_weights,
                                    self._rnd_state)

            destroyed = self.destroy_operators[d_idx](current, self._rnd_state)
            candidate = self.repair_operators[r_idx](destroyed, self._rnd_state)

            current, weight = self._consider_candidate(best, current, candidate,
                                                       weights, criterion)

            if current.objective() < best.objective():
                best = current

            # The weights are updated as convex combinations of the current
            # weight and the update parameter. See eq. (2), p. 12.
            d_weights[d_idx] *= operator_decay
            d_weights[d_idx] += (1 - operator_decay) * weight

            r_weights[r_idx] *= operator_decay
            r_weights[r_idx] += (1 - operator_decay) * weight

            if collect_stats:
                statistics.collect_objective(current.objective())

        return Result(best, statistics if collect_stats else None)

    def _consider_candidate(self, best, current, candidate, weights,
                            criterion):
        """
        Considers the candidate solution by comparing it against the best and
        current solutions. Returns the new solution when it is better or
        accepted, or the current in case it is rejected. Candidate solutions
        are accepted based on the passed-in acceptance criterion.

        Parameters
        ----------
        best : State
            Best solution encountered so far.
        current : State
            Current solution.
        candidate : State
            Candidate solution.
        weights : array_like
            Updating weights for the operators.
        criterion : AcceptanceCriterion
            The chosen acceptance criterion.

        Returns
        -------
        State
            The new state.
        float
            The weight value to use when updating the operator weights.
        """
        if candidate.objective() < best.objective():
            return candidate, weights[WeightIndex.IS_BEST]

        if candidate.objective() < current.objective():
            return candidate, weights[WeightIndex.IS_BETTER]

        if criterion.accept(self._rnd_state, best, current, candidate):
            return candidate, weights[WeightIndex.IS_ACCEPTED]

        return current, weights[WeightIndex.IS_REJECTED]

    def _validate_parameters(self, weights, operator_decay, iterations):
        """
        Helper method to validate the passed-in ALNS parameters.
        """
        if len(self.destroy_operators) == 0 or len(self.repair_operators) == 0:
            raise ValueError("Missing at least one destroy or repair operator.")

        if not (0 < operator_decay < 1):
            raise ValueError("Operator decay parameter outside unit interval"
                             " is not understood.")

        if any(weight <= 0 for weight in weights):
            raise ValueError("Non-positive weights are not understood.")

        if len(weights) < 4:
            # More than four is not explicitly problematic, as we only use the
            # first four anyways.
            raise ValueError("Unsupported number of weights: expected 4,"
                             " found {0}.".format(len(weights)))

        if iterations <= 0:
            raise ValueError("Non-positive number of iterations.")


def select_operator(operators, weights, rnd_state):
    """
    Selects an operator from the list of operators, using a distribution
    inferred from the given weights.

    Parameters
    ----------
    operators : array_like
        The list of operators.
    weights : array_like
        The operator weights.
    rnd_state : rnd.RandomState
        Random state to draw the choice from.

    Returns
    -------
    int
        Index into the operator array of the selected method.
    """
    return rnd_state.choice(np.arange(0, len(operators)),
                            p=weights / np.sum(weights))



import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises

from alns.criteria import RecordToRecordTravel
from alns.tests.states import One, Zero


def test_raises_negative_parameters():
    """
    Record-to-record travel does not work with negative parameters, so those
    should not be accepted.
    """
    with assert_raises(ValueError):         # start threshold cannot be
        RecordToRecordTravel(-1, 1, 1)      # negative

    with assert_raises(ValueError):         # nor can the end threshold
        RecordToRecordTravel(1, -1, 1)

    with assert_raises(ValueError):         # nor the updating step
        RecordToRecordTravel(1, 1, -1)


def test_raises_explosive_step():
    """
    For exponential updating, the step parameter must not be bigger than one,
    as that would result in an explosive threshold.
    """
    with assert_raises(ValueError):
        RecordToRecordTravel(2, 1, 2, "exponential")

    RecordToRecordTravel(2, 1, 1, "exponential")    # boundary should be fine


def test_threshold_boundary():
    """
    The boundary case for the end threshold parameter is at zero, which should
    be accepted.
    """
    RecordToRecordTravel(1, 0, 1)


def test_raises_start_smaller_than_end():
    """
    The initial threshold at the start should be bigger (or equal) to the end
    threshold.
    """
    with assert_raises(ValueError):
        RecordToRecordTravel(0, 1, 1)

    RecordToRecordTravel(1, 1, 1)           # should not raise for equality


def test_does_not_raise():
    """
    This set of parameters should work correctly.
    """
    RecordToRecordTravel(10, 5, 2)


def test_step():
    """
    Tests if the step parameter is correctly set.
    """
    for step in range(100):
        assert_equal(RecordToRecordTravel(1, 1, step).step, step)


def test_start_threshold():
    """
    Tests if the start_threshold parameter is correctly set.
    """
    for start in range(100):
        assert_equal(RecordToRecordTravel(start, 0, 1).start_threshold, start)


def test_end_threshold():
    """
    Tests if the end_threshold parameter is correctly set.
    """
    for end in range(100):
        assert_equal(RecordToRecordTravel(100, end, 1).end_threshold, end)


def test_accepts_better():
    record_travel = RecordToRecordTravel(1, 0, 0.1)
    assert_(record_travel.accept(rnd.RandomState(), One(), Zero(), Zero()))


def test_rejects_worse():
    record_travel = RecordToRecordTravel(0.5, 0.2, 0.1)

    # This results in a relative worsening of plus one, which is bigger than
    # the threshold (0.5).
    assert_(not record_travel.accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_accepts_equal():
    record_travel = RecordToRecordTravel(0, 0, 0.1)

    # Even a the strictest threshold, this should be accepted since the
    # relative improvement is zero (they are equal).
    assert_(record_travel.accept(rnd.RandomState(), Zero(), Zero(), Zero()))


def test_linear_threshold_update():
    record_travel = RecordToRecordTravel(5, 0, 1)

    # For the first five, the threshold is, resp., 5, 4, 3, 2, 1. The relative
    # worsening is plus one, so this should be accepted (lower or equal to
    # threshold).
    for _ in range(5):
        assert_(record_travel.accept(rnd.RandomState(), Zero(), Zero(), One()))

    # Threshold is now zero, so this should no longer be accepted.
    assert_(not record_travel.accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_exponential_threshold_update():
    record_travel = RecordToRecordTravel(5, 0, 0.1, "exponential")

    # The relative worsening is plus one, so this should be accepted initially,
    # as the threshold is 5, resp. 0.5. In the second case, 1 > 0.5, so the
    # second should be rejected.
    assert_(record_travel.accept(rnd.RandomState(), Zero(), Zero(), One()))
    assert_(not record_travel.accept(rnd.RandomState(), Zero(), Zero(), One()))




import numpy as np


class Statistics:

    def __init__(self):
        """
        Statistics object that stores some iteration results, which is
        optionally populated by the ALNS algorithm.
        """
        self._objectives = []

    @property
    def objectives(self):
        """
        Returns an array of previous objective values, tracking progress.
        """
        return np.array(self._objectives)

    def collect_objective(self, objective):
        """
        Collects an objective value.

        Parameters
        ----------
        objective : float
            The objective value to be collected.
        """
        self._objectives.append(objective)



import numpy.random as rnd
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from alns import ALNS, State
from alns.criteria import HillClimbing, SimulatedAnnealing
from .states import One, Zero


# HELPERS ----------------------------------------------------------------------


def get_alns_instance(repair_operators=None, destroy_operators=None, seed=None):
    """
    Test helper method.
    """
    alns = ALNS(rnd.RandomState(seed))

    if repair_operators is not None:
        for repair_operator in repair_operators:
            alns.add_repair_operator(repair_operator)

    if destroy_operators is not None:
        for destroy_operator in destroy_operators:
            alns.add_destroy_operator(destroy_operator)

    return alns


class ValueState(State):
    """
    Helper state for testing random values.
    """

    def __init__(self, value):
        self._value = value

    def objective(self):
        return self._value


# OPERATORS --------------------------------------------------------------------


def test_add_destroy_operator():
    """
    Tests if adding a destroy operator correctly updates the number of
    operators available on the ALNS instance.
    """
    alns = ALNS()

    for count in [1, 2]:
        alns.add_destroy_operator(lambda state, rnd: None)
        assert_equal(len(alns.destroy_operators), count)


def test_add_repair_operator():
    """
    Tests if adding a repair operator correctly updates the number of
    operators available on the ALNS instance.
    """
    alns = ALNS()

    for count in [1, 2]:
        alns.add_repair_operator(lambda state, rnd: None)
        assert_equal(len(alns.repair_operators), count)


# PARAMETERS -------------------------------------------------------------------


def test_raises_missing_destroy_operator():
    """
    Tests if the algorithm raises when no destroy operators have been set.
    """
    alns = get_alns_instance([lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], 0.95, HillClimbing())


def test_raises_missing_repair_operator():
    """
    Tests if the algorithm raises when no repair operators have been set.
    """
    alns = get_alns_instance(destroy_operators=[lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], 0.95, HillClimbing())


def test_raises_negative_operator_decay():
    """
    Tests if the algorithm raises when a negative operator decay parameter is
    passed.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], -0.5, HillClimbing())


def test_raises_explosive_operator_decay():
    """
    Tests if the algorithm raises when an explosive operator decay parameter is
    passed.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], 1.2, HillClimbing())


def test_raises_boundary_operator_decay():
    """
    The boundary cases, zero and one, should both raise.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], 0, HillClimbing())

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], 1, HillClimbing())


def test_raises_insufficient_weights():
    """
    We need (at least) four weights to be passed-in, one for each updating
    scenario.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1], .5, HillClimbing())


def test_raises_non_positive_weights():
    """
    The passed-in weights should all be strictly positive.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 0, 1], .5, HillClimbing())

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, -5, 1], .5, HillClimbing())


def test_raises_non_positive_iterations():
    """
    The number of iterations should be strictly positive.
    """
    alns = get_alns_instance([lambda state, rnd: None],
                             [lambda state, rnd: None])

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], .5, HillClimbing(), 0)

    with assert_raises(ValueError):
        alns.iterate(One(), [1, 1, 1, 1], .5, HillClimbing(), -1)


def test_does_not_raise():
    """
    This set of parameters, on the other hand, should work correctly.
    """
    alns = get_alns_instance([lambda state, rnd: One()],
                             [lambda state, rnd: One()])

    alns.iterate(Zero(), [1, 1, 1, 1], .5, HillClimbing(), 100)


# EXAMPLES ---------------------------------------------------------------------


def test_trivial_example():
    """
    This tests the ALNS algorithm on a trivial example, where the initial
    solution is zero, and any other operator returns one.
    """
    alns = get_alns_instance([lambda state, rnd: Zero()],
                             [lambda state, rnd: Zero()])

    result = alns.iterate(One(), [1, 1, 1, 1], .5, HillClimbing(), 100)

    assert_equal(result.best_state.objective(), 0)


def test_fixed_seed_outcomes():
    """
    Tests if fixing a seed results in deterministic outcomes even when using a
    'random' acceptance criterion (here SA).
    """
    outcomes = [0.00469, 0.00011, 0.04312]

    for seed, desired in enumerate(outcomes):                   # idx is seed
        alns = get_alns_instance(
            [lambda state, rnd: ValueState(rnd.random_sample())],
            [lambda state, rnd: None],
            seed)

        simulated_annealing = SimulatedAnnealing(1, .25, 1 / 100)

        result = alns.iterate(One(), [1, 1, 1, 1], .5, simulated_annealing, 100)

        assert_almost_equal(result.best_state.objective(), desired, decimal=5)


# TODO test more complicated examples?
# TODO test select_operator?



from numpy.testing import assert_

from alns.Statistics import Statistics


def test_empty_new_statistics():
    """
    Tests if a new Statistics object starts empty.
    """
    statistics = Statistics()
    assert_(len(statistics.objectives) == 0)


def test_collect_objectives():
    """
    Tests if a Statistics object properly collects objective values.
    """
    statistics = Statistics()

    for objective in range(1, 100):
        statistics.collect_objective(objective)

        assert_(len(statistics.objectives) == objective)
        assert_(statistics.objectives[-1] == objective)



import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises

from alns.criteria import SimulatedAnnealing
from alns.tests.states import One, Zero


def test_raises_negative_parameters():
    """
    Simulated annealing does not work with negative parameters, so those should
    not be accepted.
    """
    with assert_raises(ValueError):         # start temperature cannot be
        SimulatedAnnealing(-1, 1, 1)        # negative

    with assert_raises(ValueError):         # nor can the end temperature
        SimulatedAnnealing(1, -1, 1)

    with assert_raises(ValueError):         # nor the updating step
        SimulatedAnnealing(1, 1, -1)


def test_raises_explosive_step():
    """
    For exponential updating, the step parameter must not be bigger than one,
    as that would result in an explosive threshold.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(2, 1, 2, "exponential")

    SimulatedAnnealing(2, 1, 1, "exponential")    # boundary should be fine


def test_temperature_boundary():
    """
    The boundary case for the end temperature parameter is at zero, which must
    *not* be accepted, as it would result in a division-by-zero.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(1, 0, 1)


def test_raises_start_smaller_than_end():
    """
    The initial temperature at the start should be bigger (or equal) to the end
    temperature.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(0.5, 1, 1)

    SimulatedAnnealing(1, 1, 1)           # should not raise for equality


def test_does_not_raise():
    """
    This set of parameters should work correctly.
    """
    SimulatedAnnealing(10, 5, 2)


def test_step():
    """
    Tests if the step parameter is correctly set.
    """
    for step in range(100):
        assert_equal(SimulatedAnnealing(1, 1, step).step, step)


def test_start_temperature():
    """
    Tests if the start_temperature parameter is correctly set.
    """
    for start in range(1, 100):
        assert_equal(SimulatedAnnealing(start, 1, 1).start_temperature, start)


def test_end_temperature():
    """
    Tests if the end_temperature parameter is correctly set.
    """
    for end in range(1, 100):
        assert_equal(SimulatedAnnealing(100, end, 1).end_temperature, end)


def test_accepts_better():
    for _ in range(1, 100):
        simulated_annealing = SimulatedAnnealing(2, 1, 1)

        assert_(simulated_annealing.accept(rnd.RandomState(),
                                           One(),
                                           Zero(),
                                           Zero()))


def test_accepts_equal():
    simulated_annealing = SimulatedAnnealing(2, 1, 1)

    for _ in range(100):
        # This results in an acceptance probability of exp{0}, that is, one.
        # Thus, the candidate state should always be accepted.
        assert_(simulated_annealing.accept(rnd.RandomState(),
                                           One(),
                                           One(),
                                           One()))


def test_linear_random_solutions():
    """
    Checks if the linear ``accept`` method correctly decides in two known cases
    for a fixed seed.
    """
    simulated_annealing = SimulatedAnnealing(2, 1, 1)

    state = rnd.RandomState(0)

    # Using the above seed, the first two random numbers are 0.55 and .72,
    # respectively. The acceptance probability is 0.61 first, so the first
    # should be accepted (0.61 > 0.55). Thereafter, it drops to 0.37, so the
    # second should not (0.37 < 0.72).
    assert_(simulated_annealing.accept(state, Zero(), Zero(), One()))
    assert_(not simulated_annealing.accept(state, Zero(), Zero(), One()))


def test_exponential_random_solutions():
    """
    Checks if the exponential ``accept`` method correctly decides in two known
    cases for a fixed seed. This is the exponential equivalent to the linear
    random solutions test above.
    """
    simulated_annealing = SimulatedAnnealing(2, 1, 0.5, "exponential")

    state = rnd.RandomState(0)

    assert_(simulated_annealing.accept(state, Zero(), Zero(), One()))
    assert_(not simulated_annealing.accept(state, Zero(), Zero(), One()))




import sys

import pytest
from numpy.testing import assert_, assert_raises

from alns.Result import Result
from alns.Statistics import Statistics
from alns.exceptions import NotCollectedError
from .states import Sentinel

try:
    from matplotlib.testing.decorators import check_figures_equal
except ImportError:
    def check_figures_equal(*args, **kwargs):       # placeholder
        return check_figures_equal


# HELPERS ----------------------------------------------------------------------


def get_statistics():
    """
    Helper method.
    """
    statistics = Statistics()

    for objective in range(100):
        statistics.collect_objective(objective)

    return statistics


def get_plot(ax, *args, **kwargs):
    """
    Helper method.
    """
    ax.plot(*args, **kwargs)

    ax.set_title("Objective value at each iteration")
    ax.set_ylabel("Objective value")
    ax.set_xlabel("Iteration (#)")


# TESTS ------------------------------------------------------------------------


def test_result_state():
    """
    Tests if the result object correctly returns the passed-in state.
    """
    best = Sentinel()
    result = Result(best)

    assert_(result.best_state is best)


def test_raises_missing_statistics():
    """
    Accessing the statistics object when no statistics have been passed-in
    should raise.
    """
    result = Result(Sentinel())

    with assert_raises(NotCollectedError):
        result.statistics  # pylint: disable=pointless-statement

    result = Result(Sentinel(), Statistics())
    result.statistics  # pylint: disable=pointless-statement


@pytest.mark.skipif(sys.version_info < (3, 5),
                    reason="Plot testing is not reliably available for Py3.4")
@check_figures_equal(extensions=['png'])
def test_plot_objectives(fig_test, fig_ref):
    """
    Tests if the ``plot_objectives`` method returns the same figure as a
    reference plot below.
    """
    statistics = get_statistics()
    result = Result(Sentinel(), statistics)

    # Tested plot
    result.plot_objectives(fig_test.subplots())

    # Reference plot
    get_plot(fig_ref.subplots(), statistics.objectives)


@pytest.mark.skipif(sys.version_info < (3, 5),
                    reason="Plot testing is not reliably available for Py3.4")
@check_figures_equal(extensions=['png'])
def test_plot_objectives_kwargs(fig_test, fig_ref):
    """
    Tests if the passed-in keyword arguments to ``plot_objectives`` are
    correctly passed to the ``plot`` method.
    """
    statistics = get_statistics()
    result = Result(Sentinel(), statistics)

    kwargs = dict(lw=5, marker='*')

    # Tested plot
    result.plot_objectives(fig_test.subplots(), **kwargs)

    # Reference plot
    get_plot(fig_ref.subplots(), statistics.objectives, **kwargs)


@pytest.mark.skipif(sys.version_info < (3, 5),
                    reason="Plot testing is not reliably available for Py3.4")
def test_plot_objectives_default_axes():
    """
    When an axes object is not passed, the ``plot_objectives`` method should
    create a new figure and axes object.
    """
    statistics = get_statistics()

    result = Result(Sentinel(), statistics)
    result.plot_objectives()

    # TODO verify the resulting plot somehow



from abc import ABC, abstractmethod

from ..State import State  # pylint: disable=unused-import
from numpy.random import RandomState  # pylint: disable=unused-import


class AcceptanceCriterion(ABC):
    """
    Base class from which to implement an acceptance criterion.
    """

    @abstractmethod
    def accept(self, rnd, best, current, candidate):
        """
        Determines whether to accept the proposed, candidate solution based on
        this acceptance criterion and the other solution states.

        Parameters
        ----------
        rnd : RandomState
            May be used to draw random numbers from.
        best : State
            The best solution state observed so far.
        current : State
            The current solution state.
        candidate : State
            The proposed solution state.

        Returns
        -------
        bool
            Whether to accept the candidate state (True), or not (False).
        """
        return NotImplemented



from .AcceptanceCriterion import AcceptanceCriterion
from .update import update


class RecordToRecordTravel(AcceptanceCriterion):

    def __init__(self, start_threshold, end_threshold, step, method="linear"):
        """
        Record-to-record travel, using an updating threshold. The threshold is
        updated as,

        ``threshold = max(end_threshold, threshold - step)`` (linear)

        ``threshold = max(end_threshold, step * threshold)`` (exponential)

        where the initial threshold is set to ``start_threshold``.

        Parameters
        ----------
        start_threshold : float
            The initial threshold.
        end_threshold : float
            The final threshold.
        step : float
            The updating step.
        method : str
            The updating method, one of {'linear', 'exponential'}. Default
            'linear'.

        References
        ----------
        - Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
          criteria for the adaptive large neighbourhood search metaheuristic.
          *Journal of Heuristics* (2018) 24 (5): 783–815.
        - Dueck, G., Scheuer, T. Threshold accepting: A general purpose
          optimization algorithm appearing superior to simulated annealing.
          *Journal of Computational Physics* (1990) 90 (1): 161-175.
        """
        if start_threshold < 0 or end_threshold < 0 or step < 0:
            raise ValueError("Thresholds must be positive.")

        if start_threshold < end_threshold:
            raise ValueError("Start threshold must be bigger than end "
                             "threshold.")

        if method == "exponential" and step > 1:
            raise ValueError("For exponential updating, the step parameter "
                             "must not be explosive.")

        self._start_threshold = start_threshold
        self._end_threshold = end_threshold
        self._step = step
        self._method = method

        self._threshold = start_threshold

    @property
    def start_threshold(self):
        return self._start_threshold

    @property
    def end_threshold(self):
        return self._end_threshold

    @property
    def step(self):
        return self._step

    @property
    def method(self):
        return self._method

    def accept(self, rnd, best, current, candidate):
        # This follows from the paper by Dueck and Scheueur (1990), p. 162.
        result = (candidate.objective() - best.objective()) <= self._threshold

        self._threshold = max(self.end_threshold,
                              update(self._threshold, self.step, self.method))

        return result



import numpy as np

from .AcceptanceCriterion import AcceptanceCriterion
from .update import update


class SimulatedAnnealing(AcceptanceCriterion):

    def __init__(self, start_temperature, end_temperature, step,
                 method="linear"):
        """
        Simulated annealing, using an updating temperature. The temperature is
        updated as,

        ``temperature = max(end_temperature, temperature - step)`` (linear)

        ``temperature = max(end_temperature, step * temperature)`` (exponential)

        where the initial temperature is set to ``start_temperature``.

        Parameters
        ----------
        start_temperature : float
            The initial temperature.
        end_temperature : float
            The final temperature.
        step : float
            The updating step.
        method : str
            The updating method, one of {'linear', 'exponential'}. Default
            'linear'.

        References
        ----------
        - Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
          criteria for the adaptive large neighbourhood search metaheuristic.
          *Journal of Heuristics* (2018) 24 (5): 783–815.
        - Kirkpatrick, S., Gerlatt, C. D. Jr., and Vecchi, M. P. Optimization
          by Simulated Annealing. *IBM Research Report* RC 9355, 1982.
        """
        if start_temperature <= 0 or end_temperature <= 0 or step < 0:
            raise ValueError("Temperatures must be strictly positive.")

        if start_temperature < end_temperature:
            raise ValueError("Start temperature must be bigger than end "
                             "temperature.")

        if method == "exponential" and step > 1:
            raise ValueError("For exponential updating, the step parameter "
                             "must not be explosive.")

        self._start_temperature = start_temperature
        self._end_temperature = end_temperature
        self._step = step
        self._method = method

        self._temperature = start_temperature

    @property
    def start_temperature(self):
        return self._start_temperature

    @property
    def end_temperature(self):
        return self._end_temperature

    @property
    def step(self):
        return self._step

    @property
    def method(self):
        return self._method

    def accept(self, rnd, best, current, candidate):
        probability = np.exp((current.objective() - candidate.objective())
                             / self._temperature)

        # We should not set a temperature that is lower than the end
        # temperature.
        self._temperature = max(self.end_temperature, update(self._temperature,
                                                             self.step,
                                                             self.method))

        return probability >= rnd.random_sample()



import matplotlib.pyplot as plt
from matplotlib.pyplot import Axes  # pylint: disable=unused-import

from .State import State  # pylint: disable=unused-import
from .Statistics import Statistics  # pylint: disable=unused-import
from .exceptions import NotCollectedError


class Result:

    def __init__(self, best, statistics=None):
        """
        Stores ALNS results. An instance of this class is returned once the
        algorithm completes.

        Parameters
        ----------
        best : State
            The best state observed during the entire iteration.
        statistics : Statistics
            Statistics optionally collected during iteration.
        """
        self._best = best
        self._statistics = statistics

    @property
    def best_state(self):
        """
        The best state observed during the entire iteration.

        Returns
        -------
        State
            The associated State object
        """
        return self._best

    @property
    def statistics(self):
        """
        The statistics object populated during iteration.

        Raises
        ------
        NotCollectedError
            When statistics were not collected during iteration. This may be
            remedied by setting the appropriate flag.

        Returns
        -------
        Statistics
            The statistics object.
        """
        if self._statistics is None:
            raise NotCollectedError("Statistics were not collected during "
                                    "iteration.")

        return self._statistics

    def plot_objectives(self, ax=None, **kwargs):
        """
        Plots the collected objective values at each iteration.

        Parameters
        ----------
        ax : Axes
            Optional axes argument. If not passed, a new figure and axes are
            constructed.
        kwargs : dict
            Optional arguments passed to ``ax.plot``.
        """
        if ax is None:
            _, ax = plt.subplots()

        ax.plot(self.statistics.objectives, **kwargs)

        ax.set_title("Objective value at each iteration")
        ax.set_ylabel("Objective value")
        ax.set_xlabel("Iteration (#)")

        plt.draw_if_interactive()



from alns import State


class Zero(State):
    """
    Testable state with objective zero.
    """

    def objective(self):
        return 0


class One(State):
    """
    Testable state with objective one.
    """

    def objective(self):
        return 1


class Two(State):
    """
    Testable state with objective two.
    """

    def objective(self):
        return 2


class Sentinel(Zero):
    """
    Placeholder state.
    """
    pass



import setuptools

MAJOR = 1
MINOR = 0
MAINTENANCE = 4
MODIFIER = ""

VERSION = "{0}.{1}.{2}{3}".format(MAJOR, MINOR, MAINTENANCE, MODIFIER)


# Long description is simply the README file
with open("README.md", "r") as file:
    long_description = file.read()

setuptools.setup(
    name="alns",
    version=VERSION,
    author="Niels Wouda",
    author_email="nielswouda@gmail.com",
    description="A flexible implementation of the adaptive large neighbourhood"
                " search (ALNS) algorithm.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/N-Wouda/ALNS",
    project_urls={
        "Tracker": "https://github.com/N-Wouda/ALNS/issues",
        "Source": "https://github.com/N-Wouda/ALNS",
    },
    packages=setuptools.find_packages(exclude=['examples']),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
    ],
    python_requires='~=3.4',
    install_requires=[
        'numpy >= 1.15.2',
        'matplotlib >=  2.2.0',
    ])



from enum import IntEnum, unique


@unique
class WeightIndex(IntEnum):
    """
    Names the various indices in the weights list. See eq. (1), p. 12 in
    Pisinger and Røpke (2010).
    """
    IS_BEST = 0
    IS_BETTER = 1
    IS_ACCEPTED = 2
    IS_REJECTED = 3



from alns.criteria.update import update
from numpy.testing import assert_raises, assert_equal


def test_raises_unknown_method():
    with assert_raises(ValueError):
        update(1, 0.5, "unknown_method")

    update(1, 0.5, "linear")            # this should work


def test_accepts_any_case_method():
    """
    ``update`` should be indifferent about the passed-in method casing.
    """
    assert_equal(update(1, 0.5, "linear"),
                 update(1, 0.5, "LINEAR"))

    assert_equal(update(1, 0.5, "exponential"),
                 update(1, 0.5, "EXPONENTIAL"))


def test_linear():
    """
    Tests if linear updating works as designed, as ``current - step``.
    """
    assert_equal(update(1, 0.5, "linear"), 0.5)
    assert_equal(update(2, 1, "linear"), 1)
    assert_equal(update(2, 0, "linear"), 2)


def test_exponential():
    """
    Tests if exponential updating works as designed, as ``current * step``.
    """
    assert_equal(update(1, 0.5, "exponential"), 0.5)
    assert_equal(update(2, 1, "exponential"), 2)
    assert_equal(update(2, 0, "exponential"), 0)
